
function check_rows_length(x::AbstractMatrix, raxis)
    length(raxis) == size(x, 1) || error("All columns must have the same number of rows.")
    return nothing
end

function check_cols_length(x::AbstractMatrix, caxis)
    length(caxis) == size(x, 2) || error("Column axis must have the same length as the number of columns.")
    return nothing
end

function check_rows_length(x::AbstractVector{<:AbstractVector}, raxis)
    nr = length(raxis)
    for x_i in x
        length(x_i) == nr || error("All columns must have the same number of rows.")
    end
    return nothing
end

function check_cols_length(x::AbstractVector{<:AbstractVector}, caxis)
    length(caxis) == length(x) || error("Column axis must have the same length as the number of columns.")
    return nothing
end

check_cols_type(x::AbstractVector{<:AbstractVector}, caxis::AbstractAxis) = nothing

function check_cols_type(x::AbstractVector{<:AbstractVector}, caxis::StructAxis)
    @inbounds for i in axes(x)
        eltype(x[i]) <: axis_eltype(caxis, i) || error("Specified eltypes and column eltypes don't match.")
    end
    return nothing
end

function _create_col_axis(x::AbstractVector{<:AbstractVector}, caxis::AbstractAxis)
    check_cols_length(x, caxis)
    return caxis
end

function _create_col_axis(x::AbstractMatrix, caxis::AbstractAxis)
    check_cols_length(x, caxis)
    return caxis
end

function _create_col_axis(x::AbstractMatrix, caxis::AbstractVector{Symbol})
    return to_axis(caxis, indices(x, 2))
end

function _create_col_axis(x::AbstractVector{<:AbstractVector}, caxis::AbstractVector{Symbol})
    return to_axis(caxis, indices(x, 1))
end

function _create_col_axis(x::AbstractVector{<:AbstractVector}, ::Type{T}) where {T}
    return StructAxis{T}(indices(x, 1))
end

_create_col_axis(x::AbstractMatrix, ::Nothing) = _create_col_axis(x, col_axis(x))
function _create_col_axis(x::AbstractVector{<:AbstractVector}, ::Nothing)
    return _create_col_axis(x, row_axis(x))
end

function _create_col_axis(x::AbstractMatrix, caxis::AbstractVector{<:Integer})
    return to_axis([Symbol(:x, col_i) for col_i in caxis], indices(x, 1))
end
function _create_col_axis(x::AbstractVector{<:AbstractVector}, caxis::AbstractVector{<:Integer})
    return to_axis([Symbol(:x, col_i) for col_i in caxis], indices(x, 1))
end

_create_row_axis(x::AbstractMatrix, ::Nothing) = to_axis(axes(x, 1))
function _create_row_axis(x::AbstractVector{<:AbstractVector}, ::Nothing)
    return to_axis(axes(first(x), 1))
end

function _create_row_axis(x::AbstractMatrix, raxis::AbstractVector)
    check_rows_length(x, raxis)
    return to_axis(raxis, axes(x, 1))
end
function _create_row_axis(x::AbstractVector{<:AbstractVector},  raxis::AbstractVector)
    check_rows_length(x, raxis)
    return to_axis(raxis, axes(first(x), 1))
end

function _create_row_axis(x::AbstractVector{<:AbstractVector}, raxis::AbstractAxis)
    check_rows_length(x, raxis)
    return raxis
end
function _create_row_axis(x::AbstractMatrix, raxis::AbstractAxis)
    check_rows_length(x, raxis)
    return raxis
end


"""
    Table

Stores a vector of columns that may be acccessed via the Tables.jl interface.
"""
struct Table{P<:AbstractVector{<:AbstractVector},RA,CA} <: AbstractTable{P,RA,CA}
    parent::P
    row_axis::RA
    col_axis::CA

    function Table{P,RA,CA}(x::P, raxis::RA, caxis::CA) where {P,RA,CA}
        check_cols_type(x, caxis)
        return new{P,RA,CA}(x, raxis, caxis)
    end


    function Table(x::AbstractVector{<:AbstractVector}, raxis::AbstractAxis, caxis::AbstractAxis)
        return Table{typeof(x),typeof(raxis),typeof(caxis)}(x, raxis, caxis)
    end

    function Table(x::AbstractMatrix{T}; row_keys=nothing, col_keys=nothing) where {T}
        caxis = _create_col_axis(x, col_keys)
        raxis = _create_row_axis(x, row_keys)
        data = Vector{Vector{T}}(undef, size(x, 2))
        @inbounds for i in axes(x, 2)
            data[i] = x[:,i]
        end
        return Table{typeof(data),typeof(raxis),typeof(caxis)}(data, raxis, caxis)
    end

    function Table(x::AbstractVector{<:AbstractVector}; row_keys=nothing, col_keys=nothing)
        caxis = _create_col_axis(x, col_keys)
        raxis = _create_row_axis(x, row_keys)
        return Table{typeof(x),typeof(raxis),typeof(caxis)}(x, raxis, caxis)
    end

    function Table(x::AxisVector{<:Any,<:AbstractVector{<:AbstractVector}}; row_keys=nothing, col_keys=nothing)
        caxis = _create_col_axis(x, col_keys)
        raxis = _create_row_axis(x, row_keys)
        return Table{parent_type(x),typeof(raxis),typeof(caxis)}(parent(x), raxis, caxis)
    end

    function Table(x::AxisArray{<:AbstractVector,1})
        return Table(parent(x), to_axis(axes(first(x), 1)), axes(x, 1))
    end

    Table(data::NamedTuple) = Table([d for d in data], [keys(data)...])

    function Table(data::NamedTuple{(),Tuple{}})
        Table{Vector{AbstractVector},
              SimpleAxis{Int,OneToMRange{Int}},
              Axis{Symbol,Int,Vector{Symbol},OneToMRange{Int}}
        }(
            AbstractVector[],
            SimpleAxis(OneToMRange(0)),
            Axis(Symbol[], OneToMRange(0))
        )
    end

    Table(; kwargs...) = Table(values(kwargs))

    function Table(data::AbstractDict{K,<:AbstractVector}) where {K}
        ks = K[]
        new_data = AbstractVector[]
        for (k,v) in data
            push!(ks, k)
            push!(new_data, v)
        end
        caxis = Axis(ks)
        raxis = _create_row_axis(new_data, nothing)
        return Table{typeof(new_data),typeof(raxis),typeof(caxis)}(new_data, raxis, caxis)
    end

    function Table(data::Pair{K,<:AbstractVector}...) where {K}
        ks = K[]
        new_data = AbstractVector[]
        for p_i in data
            k, v = p_i
            push!(ks, k)
            push!(new_data, v)
        end
        caxis = Axis(ks)
        raxis = _create_row_axis(new_data, nothing)
        return Table{typeof(new_data),typeof(raxis),typeof(caxis)}(new_data, raxis, caxis)
    end

    Table(table) = Table(TableTraitsUtils.create_columns_from_iterabletable(table)...)

end

Base.parent(x::Table) = getfield(x, :parent)

Interface.row_axis(x::Table) = getfield(x, :row_axis)

Interface.col_axis(x::Table) = getfield(x, :col_axis)

Base.getproperty(x::Table, i) = getindex(x, :, i)

Base.getproperty(x::Table, i::Symbol) = getindex(x, :, i)

Base.setproperty!(x::Table, i::Symbol, val) = setindex!(x, val, :, i)

Base.propertynames(x::Table) = col_keys(x)

Table(x::AbstractVector{<:AbstractVector}, ks::AbstractVector) = Table(AxisArray(x, ks))

Tables.materializer(x::Table) = Table


